#include "baseconvert.h"


int main(int argc,char **argv){
  int opt;				// tempvariable to go through the option array
  int option_index = 0;

  double in_d;			// input by user
  //int    in_int;

  format = &s_prec;		// default of outputformat = single

  struct out_bits output;
  out_int[32] = '\0';

  // struct for possible options -> getopt.h
  static struct option long_options[] = {
	{"format",1,0,'f'},
	{"reverse",1,0,'r'},
	{"signed",0,0,'s'},
	{"mantissa",1,0,'m'},
	{"exponent",1,0,'e'},
	{"help",0,0,'h'},
	{0, 0, 0, 0}  // According to the Docu  -> "The last element of the array has to be filled with zeroes. "
  };

  /* Run through all available options */
  while ((opt = getopt_long (argc, argv, "f:s?r?m:e:h",long_options, &option_index))!= -1) {
	switch( opt ){
	case 'f':
	  if(strcmp(optarg,"single")==0){
		printf("single:: %s \n",optarg);
		format = &s_prec;
	  }
	  else if(strcmp(optarg,"user")==0){
		in_format = user;
		format = &user_prec;
	  }
	  else if(strcmp(optarg,"integer")==0){
		in_format = integer;
		format = &user_prec;
	  }
	  else{
		printf("-f switch given but no valid format \n ");
	  }
	  break;
	case 'r':
	  current_mode = reverse;		 
	  break;
	case 's':
	  format->sign_bit=1;
	  sign_mode=with_sign;
	  break;
	case 'm':		
	  format->mant=atoi(optarg);
	  break;
	case 'e':		
	  format->exponent=atoi(optarg);
	  break;
	case 'h':
	  printhelp();
	  break;
	default:
	  printhelp();
	  break;
	}
  }	

  /* Error Checking */
  if(in_format == user){
	if(format->exponent == 0 || format->mant ==0){
	  printf("ERROR:\nBoth bits for exponent and mantissa have to be greater than 0\n you specified:");
	  printf("mantissa: %d exponent %d \n",format->mant,format->exponent);
	  exit(1);
	}
  }  
	
  /* get the memory we need for output */
  output.sign_bit = (char *)malloc(format->sign_bit+1);
  output.exponent = (char *)malloc(format->exponent+1);
  output.mant= (char *)malloc(format->mant+1);
  bzero(output.sign_bit,format->sign_bit+1);
  bzero(output.mant,format->mant+1);
  bzero(output.exponent,format->exponent+1);

  if(current_mode ==forward){
	if(in_format == integer){
	  printf("Enter integer to convert:\n");
	  scanf("%lf",&in_d);
	  int2bin(in_d,out_int,32);
	  prettyprint_int(out_int,32);
	}
	else{
	  printf("Enter float to convert:\n");
	  scanf("%lf",&in_d);
	  double2bin(in_d, &output , format);
	  prettyprint_struct(&output,format);
	}
  }
  else if(current_mode == reverse ) {
	printf("Enter bitstring to convert (32-Bit max) leading bit is sign:\n");
	scanf("%s",in_string);
	if(in_format == integer){
	  int out = bin2int(in_string);
	  printf("Integer: %d Hex: 0x%x \n",out,out );
	}
	else{
	  double out= bin2double(in_string,format);
	  printf("Double: %lf \n",out);
	}
  } 
  printf("\n");
  return 0;
}



/*  convert interger to a characterarray of 0 and 1 */
/*  Arguments: in    -> number to convert			*/
/* 			out   -> chararray of [1,0]				*/
/* 		    width -> length of array				*/
void int2bin(int in, char* out,int width){
  int i,stelle;
  if( in >0){ 
	for(i=width-1; i>=0 ; i--){
	  stelle = pow(2,i);
	  if((in - stelle) >=0){  
		in -= stelle;
		*out='1';	  
	  }
	  else{
		*out='0';
	  }
	  out=out+1;
	}
  } 
  else{
	for(i=width-1; i>=0 ; i--){
	  stelle = pow(2,i);
	  if((in-stelle) >=0){  
		in -= stelle;
		*out='1';	  
	  }
	  else{
		*out='0';
	  }
	  out=out+1;
	} 
  } 
}


/*  convert double to a characterarray of 0 and 1   */
/*  Arguments: in    -> number to convert			*/
/* 			out      -> chararray of [1,0]  		*/
/* 		    format   -> target format   			*/
void double2bin(double in,struct out_bits* out, struct F_format* format){
  int    bias=0,shift=0,exp=0,i=0;
  int    minexp,maxexp;
  double frac=0;    
  printf("Format  exp: %d mant:%d sign: %d \n",format->exponent,format->mant,format->sign_bit);
  // Sign bit
  if(in <0){
	*out->sign_bit='1';
	printf("SIGN:\t 1\n");
	in=0-in;
  }
  else{
	printf("SIGN:\t 0\n");
	*out->sign_bit='0';
  }	

  // Exponent,biased
  bias = pow(2,format->exponent-1)-1;
  printf("Bias: %d\n",bias);
  maxexp = pow(2,format->exponent-1)-1;
  minexp = -pow(2,format->exponent-1);
  i = maxexp;

  while(i >= minexp){
	frac= in/pow(2,i);
	if( (frac >=1) && (frac <2) ){
	  exp = i;
	  break;
	}
	i--;
  }
  int2bin(exp+bias,out->exponent,format->exponent);
  printf("EXPONENT:\t %d  Binary:  %s \n",exp,out->exponent);
  
  // Fraction
  frac--;
  shift = frac*pow(2,format->mant);
  int2bin(shift,out->mant,format->mant);
  printf("FRACTION:\t %lf   Binary: %s \n",frac,out->mant);
}


/*  converts characterarray of 0 and 1  to int      */
/*  Arguments: in    -> number to convert			*/
/* 			out      -> chararray of [1,0]  		*/
/* 		    format   -> target format   			*/
int bin2int(char* in){
  int i,out=0;
  int len=strlen(in);
  i = len-1;
  // input will be treated as signed 
  if(sign_mode == with_sign){
	i--;			//Skip the leading bit
	in++;
	// walk through remaining bits
	while(i>=0){
	  if(*in == '0'){
		out += pow(2,i);
	  }
	  i--;  // next power of 2
	  in++;	// next char from string 
	}
	// Do a complement 2 operation if leading bit is '1'
	if(*(in-len) =='1'){
	  out = -1-out;
	}
  }

  // input will be treated as unsigned
  else{
	while(i>=0){
	  if(*in == '1'){
		out += pow(2,i);
	  }
	  i--;  // next power of 2
	  in++;	// next char from string 
	}
  }
  return out;
}

/*  prints a array of 0,1 to console grouped in packs of 4   */
/*  Arguments: in    -> number to convert			         */
void prettyprint_int(char* in,int width){
  int i;
  int maxlen = (strlen(in)>width) ? strlen(in) : width; 
  for(i=0;i<=maxlen;i++){
	if(i!=0 && i%4 == 0){
	  putchar(' ');
	}
	printf("%c",*(in++));
  }
  i--;
}

/*  prints a array of 0,1 to console grouped in packs of 4   */
/*  Arguments: in    -> struct of sign,exp,	mantissa         */
void prettyprint_struct(struct out_bits *in ,struct F_format *format){
  int memsize=format->sign_bit+format->exponent+format->mant+1;
  char* str = (char *)malloc(memsize);
  bzero(str,memsize);
  strcat(str,in->sign_bit);
  strcat(str,in->exponent);
  strcat(str,in->mant);
  printf("Double: %s  \n",str);
  prettyprint_int(str,memsize-1);
  printf("\n");
}

/*  convert  a characterarray of 0 and 1 to double  */
/*  Arguments: in    -> number to convert			*/
/* 			out      -> chararray of [1,0]  		*/
/* 		    format   -> target format   			*/
/* 	  returns: double  corresponding to bitstring   */   
double bin2double(char* in, struct F_format *format){
  int memsize=format->sign_bit+format->exponent+format->mant+1;
  int offset;
  double exp,mant;
  double out = 0.0;
  char* str = (char *)malloc(memsize);
  bzero(str,memsize);
  offset = format->sign_bit+format->exponent;
  strcpy(str,in+offset);
  printf("Mantissa STRING: %s \n",str);
  mant = bin2int(str);
  printf("mant, no shift: %lf \n",mant);
  mant = (mant/pow(2,format->mant))+1.0;
  printf("mant, shift: %lf \n",mant);
  bzero(str,memsize);
  offset = format->sign_bit;
  strncpy(str,in+offset,format->exponent);
  printf("EXPONENT STRING: %s \n",str);
  exp = bin2int(str);
  printf("exp biased: %lf \n",exp);
  exp = exp - (pow(2,format->exponent-1)-1); // unbiased
  printf("exp unbiased: %lf \n",exp);
  out = mant*pow(2,exp);
  // sign correction
  if(*in == '1'){
	out = -1*out;
  }
  return out;
}


/* prints a quick helpreference to stdout */
void printhelp(){
      printf("Binconvert Copyright (c) 2005 Thomas Ruschvial\n");
	  printf("\tcompiled %s %s \n",__TIME__,__DATE__);
	  printf("Converts a floating point or integer to a bitstring and vice versa\n");  
	  printf("Usage: baseconvert [-r -f <format>] | [-f <format>] |  \n");
	  printf("			| [-s] [-m <int>] [-e <int>]");
	  printf("\noptional Arguments:\n");
	  printf("-f --format  \t: [single,user,integer] default is single \n");
	  printf("-r --reverse \t: string of [0,1] to single or int \n");
	  printf(" If format \'user\' is selected following format options are required: \n");
	  printf("-s --signed  \t: signed or unsigned format \n");
	  printf("-m --mantissa\t: bits for manissa \n");
	  printf("-e --exponent\t: bits for exponent\n");			  
	  exit(0);
}
